#define _GNU_SOURCE
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fnmatch.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "stdrusty.h"
#include "ccontrol.h"

static void insert_arg(char *argv[], unsigned argc, unsigned pos, char *arg)
{
	memmove(argv+pos+1, argv+pos, (argc+1) * sizeof(char *));
	argv[pos] = arg;
}

void __verbose(const char *msg, ...)
{
	va_list arglist;
	char *str, pidstr[CHAR_SIZE(int)];

	int_to_string(pidstr, getpid());
	write(STDERR_FILENO, "ccontrol: ", strlen("ccontrol: "));
	write(STDERR_FILENO, pidstr, strlen(pidstr));
	write(STDERR_FILENO, ": ", strlen(": "));
	write(STDERR_FILENO, msg, strlen(msg));

	va_start(arglist, msg);
	while ((str = va_arg(arglist, char *)) != NULL)
		write(STDERR_FILENO, str, strlen(str));
	va_end(arglist);

	write(STDERR_FILENO, "\n", 1);
}

static bool make_target_match(char *argv[], char *targets)
{
	char *p, *target = "";
	unsigned int i;

	if (!targets)
		return false;

	/* Heuristic: choose last arg which is neither option nor variable */
	for (i = 0; argv[i]; i++) {
		if (argv[i][0] == '-')
			continue;
		if (strchr(argv[i], '='))
			continue;
		target = argv[i];
	}

	for (p = strtok(targets, " \t"); p; p = strtok(NULL, " \t"))
		if (fnmatch(p, target, 0) == 0)
			return true;
	return false;
}

/* Earlier arg parsing might have revealed distcc is useless. */
static void adjust_args_and_environment(struct section sec,
					enum type type,
					char *argv[], unsigned argc)
{
	argv[0] = sec.names[type];

	switch (type) {
	case TYPE_MAKE:
		/* Ensure make calls us for children, not /usr/bin/make. */
		setenv("MAKE", "make", 1);

		/* Children on non-parallel'ed makes are not parallel. */
		if (getenv("CCONTROL_NO_PARALLEL"))
			break;

		if (make_target_match(argv, sec.no_parallel))
			setenv("CCONTROL_NO_PARALLEL", "1", 1);
		else {
			verbose(sec, "Inserting arg -j in make!", NULL);
			insert_arg(argv, argc++, 1, "-j");
		}
		break;
	case TYPE_LD:
		break;
	case TYPE_CC:
	case TYPE_CPLUSPLUS:
		/* ccache + distcc needs special care. */
		if (sec.distcc) {
			verbose(sec, "Setting DISTCC_HOSTS=", 
				sec.distcc_hosts, NULL);
			setenv("DISTCC_HOSTS", sec.distcc_hosts, 1);
			if (sec.ccache) {
				verbose(sec, "Setting CCACHE_PREFIX=",
					sec.distcc, NULL);
				setenv("CCACHE_PREFIX", sec.distcc, 1);
			} else {
				verbose(sec, "Prefixing arg ",
					sec.distcc, NULL);
				insert_arg(argv, argc++, 0, sec.distcc);
			}
		}

		/* Don't set CCACHE_UNIFY: misses not so bad with distcc. */
		if (sec.ccache) {
			verbose(sec, "Prefixing arg ", sec.ccache, NULL);
			insert_arg(argv, argc++, 0, sec.ccache);
		}
	}
}

static void remove_distcc(struct section sec, char *argv[], int size)
{
	if (sec.ccache)
		unsetenv("CCACHE_PREFIX");
	else
		/* First arg is distcc. */
		memmove(argv, argv+1, size - sizeof(argv[0]));
}

#ifndef TESTING
static void __attribute__((noreturn)) run_command(struct section sec,
						  char *argv[])
{
	verbose(sec, "Execing '", argv[0], "'", NULL);
	execv(argv[0], argv);
	fatal("failed to exec '", errno, argv[0], "'", NULL);
}
#else
static void __attribute__((noreturn)) run_command(struct section sec,
						  char *argv[])
{
	if (getenv("DISTCC_HOSTS"))
		printf("DISTCC_HOSTS='%s' ", getenv("DISTCC_HOSTS"));

	while (argv[0]) {
		printf("%s ", argv[0]);
		argv++;
	}
	printf("\n");

	exit(atoi(getenv("EXITCODE") ?: "0"));
}
#endif

/* Fork command, wait for it, return exit status. */
static int fork_command(struct section sec, char *argv[])
{
	int status, pid = fork();
	char num[20];

	if (pid < 0)
		fatal("failed to fork", errno, NULL);
	if (pid == 0)
		run_command(sec, argv);
	int_to_string(num, pid);
	verbose(sec, "Forked ", num, NULL);
	waitpid(pid, &status, 0);
	int_to_string(num, WIFEXITED(status) ? WEXITSTATUS(status) : 255);
	verbose(sec, "Waitpid returned ", num, NULL);
	return WIFEXITED(status) ? WEXITSTATUS(status) : 255;
}

int main(int argc, char *orig_argv[])
{
	enum type type;
	struct section sec;
	int lockfd, ret;
	char configname[PATH_MAX];
	char *argv[argc + 3];

	/* Run low priority; people like to use apps while compiling. */
	nice(10);

	strcpy(configname, getenv("HOME"));
	strcat(configname, "/.ccontrol/default");

	/* Make room to add args. */
	memcpy(argv, orig_argv, (argc + 1)*sizeof(argv[0]));

#ifdef TESTING
	if (getenv("ARGV0"))
		argv[0] = getenv("ARGV0");
#endif
	sec = read_config(configname);
	type = what_am_i(argv);

	/* Check we really distcc if asked to. */
	if (sec.distcc) {
		if (type == TYPE_LD || type == TYPE_MAKE)
			sec.distcc = NULL;
		else if (!can_distcc(argv)) {
			verbose(sec, "Cannot distcc this", NULL);
			sec.distcc = NULL;
		}
	}

	adjust_args_and_environment(sec, type, argv, argc);

	lockfd = grab_lock(configname, sec, type);

	if (!sec.distcc)
		run_command(sec, argv);

	/* We have to wait, to make sure distcc actually distributes,
	 * otherwise we end up running 20 compiles locally. */
	setenv("DISTCC_FALLBACK", "0", 1);

	ret = fork_command(sec, argv);
	if (ret == 103 || ret == 116) {
		char num[20];

		/* distcc failed to distribute: run single-thread. */
		remove_distcc(sec, argv, sizeof(argv));
		int_to_string(num, lockfd);
		verbose(sec, "Releasing lock fd ", num, NULL);
		close(lockfd);
		regrab_slow_lock(configname, sec);
		run_command(sec, argv);
	}
	return ret;
}
